<?php

namespace Arma\Internal;

/**
 * Interface for objects with a dense layout (Mat, Col and Row).
 *
 * @package Arma\Internal
 */
interface Dense extends Common
{
    /**
     * {@inheritdoc}
     *
     * @param int $i
     * @param int $j[optional]
     * @return \Arma\SvVal
     */
    function at($i, $j);

    /**
     * {@inheritdoc}
     *
     * @param int $i
     * @param int $j[optional]
     * @return \Arma\SvVal
     */
    function __invoke($i, $j);

    /**
     * Element-wise equality evaluation of two objects.
     *
     *
     * @param static $other
     * @return \Arma\IMat
     */
    function equals($other);

    /**
     * Element-wise non-equality evaluation of two objects.
     *
     * @param static $other
     * @return \Arma\IMat
     */
    function notEquals($other);

    /**
     * Element-wise "greater than" evaluation of two objects.
     *
     * @param static $other
     * @return \Arma\IMat
     */
    function greaterThan($other);

    /**
     * Element-wise "smaller than" evaluation of two objects.
     *
     * @param static $other
     * @return \Arma\IMat
     */
    function smallerThan($other);

    /**
     * Element-wise "not greater than" evaluation of two objects.
     *
     * @param static $other
     * @return \Arma\IMat
     */
    function notGreaterThan($other);

    /**
     * Element-wise "not smaller than" evaluation of two objects.
     *
     * @param static $other
     * @return \Arma\IMat
     */
    function notSmallerThan($other);

    /**
     * Sets the elements to a specified value. The type of value must match the type of elements used by
     * the container object.
     *
     * For matrices, filling is done column-by-column.
     *
     * @param number|\Arma\Complex $value
     * @return void
     */
    function fill($value);

    /**
     * Fill with values provided by a callback function.
     *
     * For matrices, filling is done column-by-column.
     *
     * @param callable $callback () => $fill_value
     * @return void
     */
    function imbue($callback);
}
