<?php

namespace Arma\Internal;

/**
 * Interface for vectors (Col, Row, SpCol and SpRow).
 *
 * @package Arma\Internal
 */
interface Vector
{
    // Constructors

    /**
     * Construct the vector to have user specified size and fill with specified pattern.
     *
     * @param int $n_elem[optional]
     * @param int $fill[optional]
     * @return static
     */
    static function init($n_elem, $fill);

    /**
     * Create the vector from a textual description.
     *
     * Elements should be separated by spaces.
     *
     * @param string $text
     * @return static
     */
    static function fromString($text);

    // Initialization

    /**
     * Set the elements of an object to zero, optionally first changing the size to specified dimensions.
     *
     * @param int $n_elem[optional]
     * @return void
     */
    function zeros($n_elem);

    // Vector subview

    /**
     * Read/write access to a vector subview.
     *
     * @param int $first_idx
     * @param int $last_idx
     * @return Subview
     */
    function subvec($first_idx, $last_idx);

    /**
     * Read/write access to first specified elements of a vector.
     *
     * @param int $n_elem
     * @return Subview
     */
    function head($n_elem);

    /**
     * Read/write access to last specified elements of a vector.
     *
     * @param int $n_elem
     * @return Subview
     */
    function tail($n_elem);
}
